#emacs, this is -*-Python-*- mode

cdef extern from "stdlib.h" nogil:
    ctypedef int size_t
    void *memcpy(void*,void*,size_t)
    void *memset(void*,int,size_t)
    void *malloc(size_t size)
    void free(void* ptr)
    void exit(int status)
    int printf(char *format, ...)
    ctypedef int FILE

cdef extern from "math.h" nogil:
    int abs(int i)
    double round( double f)
    double ceil( double f)
    float ceilf( float f)
    double atan2( double y, double x)
    double sqrt(double)
    int isnan(double x)
    int isinf(double x)
    double sin(double x)
    float sinf(float x)

cdef extern from "ads_wrap_system.h" nogil:
    int close(int fd)
    size_t write( int fd, void* buf, size_t count)
    
cdef extern from "stdio.h" nogil:
    size_t fwrite( void* ptr, size_t size, size_t nmemb, FILE* stream)
    



#emacs, this is -*-Python-*- mode
# $Id: c_python.pxd 1020 2006-05-01 22:32:39Z astraw $

cimport c_lib

cdef extern from "Python.h":
    #cdef object PyExc_OSError
    ctypedef int Py_intptr_t
#    ctypedef int Py_ssize_t
    void Py_BEGIN_ALLOW_THREADS()
    void Py_END_ALLOW_THREADS()
    void Py_INCREF(object)
    void Py_DECREF(object)
    
    object PyErr_SetFromErrno(object)
    
    object PyCObject_FromVoidPtr( void* cobj, void (*destr)(void *))
    object PyCObject_FromVoidPtrAndDesc( void* cobj, void* desc, void (*destr)(void *, void *))
    int PyCObject_Check(arr)
    void* PyCObject_AsVoidPtr(object)
    
    object PyLong_FromVoidPtr( void *p)
    void* PyLong_AsVoidPtr( object )
    object PyString_FromStringAndSize( char *v, int len )
    object PyString_FromString( char *v)
    int PyFile_Check(object fobj)
    c_lib.FILE* PyFile_AsFile(object fobj)
    
#    int PyObject_AsWriteBuffer( object, void **buffer, Py_ssize_t *buffer_len)
    int PyObject_AsWriteBuffer( object, void **buffer, int *buffer_len)



#emacs, this is -*-Python-*- mode

cimport c_python
cimport ipp
cimport fic

ctypedef void* fiptr # pointer to FastImage base type

cdef class Size:
    cdef ipp.IppiSize sz

cdef class Point:
    cdef ipp.IppiPoint pt

cdef class FastImageBase:
    cdef fiptr im
    cdef c_python.Py_intptr_t shape[2] # don't use, purely for making __array_struct__ without extra malloc
    cdef c_python.Py_intptr_t strides[2]
    cdef int step # int copy of strides[0], (this could be a problem where int != intptr_t)
    cdef int view
    cdef object basetype
    cdef Size imsiz
    cdef object source_data # keep reference to original image in case it goes out of scope elsewhere

    cdef FastImageBase c_roi( self, int left, int bottom, Size size)

cdef class FastImage32f(FastImageBase) # forward definition

cdef class FastImage8u(FastImageBase):
    cdef void fast_get_absdiff_put(self, FastImage8u other, FastImage8u result, Size size)
    cdef void fast_get_sub_put(self, FastImage8u other, FastImage8u result, Size size)
    cdef void fast_set_val_masked( self, ipp.Ipp8u val, FastImage8u mask, Size size)
    cdef void fast_get_32f_copy_put(self, FastImage32f result,Size size)
    cdef void fast_get_compare_int_put_greater( self, int other_int, FastImage8u dest, Size size)

cdef class FastImage32f(FastImageBase):
    cdef void fast_toself_add_weighted_8u( self, FastImage8u other, Size size, float alpha)
    cdef void fast_toself_add_weighted_32f( self, FastImage32f other, Size size, float alpha)
    cdef void fast_get_8u_copy_put(self,FastImage8u other,Size size)
    cdef void fast_toself_square(self,Size size)
    cdef void fast_get_square_put(self, FastImage32f result, Size size)
    cdef void fast_get_subtracted_put(self,FastImage32f other,FastImage32f result,Size size)
    cdef void fast_toself_multiply(self, float val, Size size)
    cdef FastImage32f create_equal_shape_empty(self, Size size)



cdef extern from "fic.h" nogil:
    ctypedef int FicStatus
    ctypedef unsigned char Fic8u
    ctypedef float Fic32f
    ctypedef double Fic64f
    ctypedef struct FiciSize:
        int width
        int height
    ctypedef int ficMomentState_64f
    FicStatus ficMomentInitAlloc_64f(ficMomentState_64f**)
    FicStatus ficMomentFree_64f( ficMomentState_64f* )

    const char* ficGetErrorString(FicStatus)

    FicStatus ficiMaxIndx_8u_C1R( Fic8u*, int, FiciSize,
                                  Fic8u*, int*, int* )
    FicStatus ficiMinIndx_8u_C1R( Fic8u*, int, FiciSize,
                                  Fic8u*, int*, int* )
    FicStatus ficiMaxIndx_32f_C1R( Fic32f*, int, FiciSize,
                                   Fic32f*, int*, int* )
    FicStatus ficiMean_8u_C1R( Fic8u*, int, FiciSize, Fic64f*)
    FicStatus ficiDotProd_8u64f_C1R( Fic8u* pSrc1, int srcStep,
                                     Fic8u* pSrc2, int src2Step,
                                     FiciSize roiSize, Fic64f* pDp )
    FicStatus ficiDotProd_32f64f_C1R( Fic32f* pSrc1, int srcStep,
                                      Fic32f* pSrc2, int src2Step,
                                      FiciSize roiSize, Fic64f* pDp )

    FicStatus ficiFilterSobelHoriz_32f_C1R( Fic32f *pSrc, int srcStep, Fic32f *pDst, int dstStep, FiciSize dstRoiSize )
    FicStatus ficiFilterSobelVert_32f_C1R ( Fic32f *pSrc, int srcStep, Fic32f *pDst, int dstStep, FiciSize dstRoiSize )
    FicStatus ficiFilterSobelHoriz_8u_C1R( Fic8u *pSrc, int srcStep, Fic8u *pDst, int dstStep, FiciSize dstRoiSize )
    FicStatus ficiFilterSobelVert_8u_C1R ( Fic8u *pSrc, int srcStep, Fic8u *pDst, int dstStep, FiciSize dstRoiSize )



#emacs, this is -*-Python-*- mode

# indexing starts from left, bottom

cimport c_lib
import traceback, sys
import numpy

cdef int FASTIMAGEDEBUG
FASTIMAGEDEBUG = 0

# create char types that work in Pyrex and Cython
cdef char CHAR_u
CHAR_u = ord('u')
cdef char CHAR_f
CHAR_f = ord('f')

# for computing histograms
cdef int n_hist_levels
n_hist_levels = 256
cdef ipp.Ipp32s hist_levels[257]
for i in range(0,257):
    hist_levels[i]=i

def set_debug(int val):
    global FASTIMAGEDEBUG
    FASTIMAGEDEBUG = val

cdef extern from "unistd.h":
    ctypedef unsigned long intptr_t

cdef extern from "limits.h":
    cdef intptr_t INT_MAX
    cdef intptr_t INT_MIN

cdef extern from "FastImage_macros.h":
    # for PyArrayInterface
    int CONTIGUOUS
    int FORTRAN
    int ALIGNED
    int NOTSWAPPED
    int WRITEABLE
    int ARR_HAS_DESCR

    # for IPP
    ipp.Ipp8u*  IMPOS8u(  ipp.Ipp8u*  im, int step, int bottom, int left)
    ipp.Ipp32f* IMPOS32f( ipp.Ipp32f* im, int step, int bottom, int left)
    void CHK_NOGIL( ipp.IppStatus errval )

    const char* ippGetStatusString( ipp.IppStatus StsCode )

    cdef void InitIPP()

    ctypedef struct ipp_version_struct_t:
        int major
        int minor
        int build

    cdef ipp_version_struct_t GetIPPVersion()
    cdef const char* GetIPPArch()
    cdef int IsIPPStatic()

InitIPP()

def get_IPP_version():
    cdef ipp_version_struct_t val
    val = GetIPPVersion()
    return val.major, val.minor, val.build

def get_IPP_arch():
    cdef const char* arch_str
    arch_str = GetIPPArch()
    return arch_str

def get_IPP_static():
    return IsIPPStatic()

class IppError(Exception):
    def __init__(self, int errval):
        cdef const char* cmsg
        cmsg = ippGetStatusString(<ipp.IppStatus>errval)
        #Exception.__init__(self,"Error %d: %s"%(errval,cmsg))
        Exception.__init__(self,cmsg)

class FicError(Exception):
    pass

cdef void CHK_HAVEGIL( ipp.IppStatus errval ) except *:
    if (errval!=0):
        raise IppError(errval)

cdef void CHK_FIC_HAVEGIL( fic.FicStatus errval ) except *:
    if (errval!=0):
        errstr = fic.ficGetErrorString(errval)
        raise FicError('error %d: %s'%(errval,errstr))

ctypedef struct PyArrayInterface:
    int two                       # contains the integer 2 as a sanity check
    int nd                        # number of dimensions
    char typekind                 # kind in array --- character code of typestr
    int itemsize                  # size of each element
    int flags                     # flags indicating how the data should be interpreted
    c_python.Py_intptr_t *shape   # A length-nd array of shape information
    c_python.Py_intptr_t *strides # A length-nd array of stride information
    void *data                    # A pointer to the first element of the array

cdef void free_array_struct( void* ptr, void *arr ):
    arrpy = <object>arr
    c_python.Py_DECREF( arrpy )
    c_lib.free(ptr)

# comparision operators
CmpLess = ipp.ippCmpLess
CmpLessEq = ipp.ippCmpLessEq
CmpEq = ipp.ippCmpEq
CmpGreaterEq = ipp.ippCmpGreaterEq
CmpGreater = ipp.ippCmpGreater

cdef class Size:
    """size of 2D image

    sz=Size(width,height)
    """
    # keep this read-only to prevent bizarre problems if multiple image share same size
    def __init__(self,int width,int height):
        self.sz.width = width
        self.sz.height = height
    def __repr__(self):
        return 'FastImage.Size(%d,%d)'%(self.sz.width,self.sz.height)
    def __richcmp__(x, y, int op):
        cdef Size xs
        cdef Size ys
        if type(x) != Size:
            raise TypeError("comparison between Size objects only")
        if type(y) != Size:
            raise TypeError("comparison between Size objects only")
        xs = x
        ys = y
        if op==2:
            return xs.sz.width == ys.sz.width and xs.sz.height == ys.sz.height
        elif op==3:
            return not (xs.sz.width == ys.sz.width and xs.sz.height == ys.sz.height)
        else:
            raise ValueError("cannot perform comparison (equality testing only)")

    # leave as short names to make life easy when looking for .sz.width (fast C) vs. .w (slow Python)
    property h:
        def __get__(self):
            return self.sz.height
    property w:
        def __get__(self):
            return self.sz.width

cdef class Point:
    # keep this read-only to prevent bizarre problems if multiple image share same origin
    def __init__(self,int x,int y):
        self.pt.x = x
        self.pt.y = y
    def __repr__(self):
        return 'FastImage.Point(%d,%d)'%(self.pt.x,self.pt.y)
    def __richcmp__(a, b, int op):
        cdef Point ap
        cdef Point bp
        if type(a) != Point:
            raise TypeError("comparison between Point objects only")
        if type(b) != Point:
            raise TypeError("comparison between Point objects only")
        ap = a
        bp = b
        if op==2:
            return ap.pt.x == bp.pt.x and ap.pt.y == bp.pt.y
        elif op==3:
            return not (ap.pt.x == bp.pt.x and ap.pt.y == bp.pt.y)
        else:
            raise ValueError("cannot perform comparison (equality testing only)")

    # leave as short names to make life easy when looking for .sz.width (fast C) vs. .w (slow Python)
    property x:
        def __get__(self):
            return self.pt.x
    property y:
        def __get__(self):
            return self.pt.y

cdef class FastImageBase:
    def __cinit__(self,*args,**kw):
        self.view = 1

    def __init__(self, Size size):
        self.imsiz = size

        # purely for making __array_struct__ without extra malloc:
        self.shape[0] = size.sz.height
        self.shape[1] = size.sz.width

    def __dealloc__(self):
        global FASTIMAGEDEBUG
        if self.view==0:
            if FASTIMAGEDEBUG>=1:
                print self,'freeing memory %#x'%<intptr_t>self.im
            ipp.ippiFree(self.im)

    def show_mem_and_step(self):
        print self,'got mem %#x step %d'%(<intptr_t>self.im,
                                          self.step)

    def stringview(self):
        cdef int i,j
        cdef fiptr rowptr
        cdef fiptr valptr
        lines = []
        for i from 0 <= i < self.imsiz.sz.height:
            line = []
            rowptr = self.im+i*self.step
            for j from 0 <= j < self.imsiz.sz.width:
                valptr = rowptr+j*self.strides[1]
                if self.basetype == '8u':
                    line.append((<ipp.Ipp8u*>valptr)[0])
                elif self.basetype == '32f':
                    line.append((<ipp.Ipp32f*>valptr)[0])
            line = ' '.join(map(str,line))
            lines.append(line)
        lines = '\n'.join(lines)
        return lines

    def dump_to_file(self,fobj):
        cdef int i
        cdef int nbytes
        cdef int nbytes_tot
        cdef int bytes_per_pixel
        cdef fiptr rowptr
        cdef fiptr valptr
        cdef c_lib.FILE* fd
        nbytes_tot = 0
        bytes_per_pixel = self.strides[1]
        if not c_python.PyFile_Check(fobj):
            raise ValueError('need file object')
        fd = c_python.PyFile_AsFile(fobj)
        with nogil:
            for i from 0 <= i < self.imsiz.sz.height:
                rowptr = self.im+i*self.step
                nbytes = c_lib.fwrite( rowptr, 1, self.imsiz.sz.width*bytes_per_pixel, fd )
                nbytes_tot = nbytes_tot+nbytes
        return nbytes_tot

    property _step:
        def __get__(self):
            return self.step

    property __array_struct__:
        def __get__(self):
            cdef PyArrayInterface* inter

            inter = <PyArrayInterface*>c_lib.malloc( sizeof( PyArrayInterface ) )
            inter.two = 2
            inter.nd = 2
            if self.basetype == '8u':
                inter.typekind = CHAR_u
            elif self.basetype == '32f':
                inter.typekind = CHAR_f
            else:
                raise ValueError('unknown basetype: %r'%self.basetype)
            inter.itemsize = self.strides[1]
            inter.flags = NOTSWAPPED | ALIGNED | WRITEABLE
            inter.strides = self.strides
            inter.shape = self.shape
            inter.data = self.im
            c_python.Py_INCREF(self)
            obj = c_python.PyCObject_FromVoidPtrAndDesc( <void*>inter, <void*>self, free_array_struct)
            return obj

    property __array_interface__:
        def __get__(self):
            inter = {}
            inter['shape'] = self.shape[0], self.shape[1]
            inter['strides'] = self.strides[0], self.strides[1]
            if self.basetype == '8u':
                inter['typestr'] = '|u1'
            else:
                raise NotImplementedError('')
            readonly = True
            data_ptr_int = <intptr_t>self.im
            inter['data'] = (data_ptr_int,readonly)
            return inter

    property size:
        def __get__(self):
            return self.imsiz

    cdef FastImageBase c_roi( self, int left, int bottom, Size size):
        cdef FastImageBase result
        global FASTIMAGEDEBUG

        if FASTIMAGEDEBUG>=2:
            print self,'creating ROI viewer'
        result = self.__new__(type(self)) # new object of my same type
        FastImageBase.__init__(result, size) # run common __init__
        result.strides[0] = self.strides[0]
        result.step = self.step
        if self.basetype == '8u':
            result.im = IMPOS8u(<ipp.Ipp8u*>self.im,self.step,bottom,left)
        elif self.basetype == '32f':
            result.im = IMPOS32f(<ipp.Ipp32f*>self.im,self.step,bottom,left)
        else:
            raise ValueError('unknown basetype: %r'%self.basetype)
        if FASTIMAGEDEBUG>=2:
            print self,'assigning ROI from self (im: %#x)\n    to (%s: 0x%0x) (L:%d,B:%d)'%(<intptr_t>self.im,
                                                          str(result),
                                                          <intptr_t>result.im,
                                                          left,bottom)
        result.source_data = self # keep reference to original image in case it goes out of scope elsewhere
        return result

    def roi( self, int left, int bottom, Size size):
        # python version of C version
        return self.c_roi(left,bottom,size)

    def __getitem__(self,a):
        cdef int i,j
        cdef fiptr valptr

        i,j = a # a must be sequence of ints

        valptr = (self.im+i*self.step)+j*self.strides[1]

        if self.basetype == '8u':
            return (<ipp.Ipp8u*>valptr)[0]
        elif self.basetype == '32f':
            return (<ipp.Ipp32f*>valptr)[0]
        else:
            raise ValueError('unknown basetype: %r'%self.basetype)

##    def __getitem__(self,*args,**kw):
##        print '*args',args
##        print '**kw',kw

cdef class FastImage32f(FastImageBase) # forward declaration
cdef class FastImage8u(FastImageBase):
    """a fast 2D array of 8u"""

    def __cinit__(self,*args,**kw):
        self.strides[1] = 1
        self.basetype = '8u'

    def __init__(self, Size size ):
        global FASTIMAGEDEBUG

        FastImageBase.__init__(self, size)
        if FASTIMAGEDEBUG>=1:
            print self,'requesting memory (size W:%d H:%d)'%(self.imsiz.sz.width,
                                                             self.imsiz.sz.height)
            if FASTIMAGEDEBUG>=3:
                sys.stderr.write('-='*20+'\n')
                traceback.print_stack()
                sys.stderr.write('-='*20+'\n\n')
                sys.stderr.flush()

        self.im=ipp.ippiMalloc_8u_C1( size.sz.width, size.sz.height, &self.step )
        if self.step<size.sz.width:
            raise MemoryError('self.step is %d (width=%d, height=%d)'%(
                self.step, size.sz.width, size.sz.height))
        if self.im==NULL: raise MemoryError("Error allocating memory")
        self.strides[0] = self.step
        if FASTIMAGEDEBUG>=1:
            self.show_mem_and_step()
        self.view = 0

    def set_val(self, int val, Size size):
        CHK_HAVEGIL( ipp.ippiSet_8u_C1R( val, <ipp.Ipp8u*>self.im, self.step, size.sz ))

    def set_val_masked(self, int val, FastImage8u mask, Size size):
        CHK_HAVEGIL( ipp.ippiSet_8u_C1MR( val, <ipp.Ipp8u*>self.im, self.step, size.sz,
                                          <ipp.Ipp8u*>mask.im, mask.step))

    cdef void fast_set_val_masked( self, ipp.Ipp8u val, FastImage8u mask, Size size):
        CHK_HAVEGIL( ipp.ippiSet_8u_C1MR( val, <ipp.Ipp8u*>self.im, self.step, size.sz,
                                        <ipp.Ipp8u*>mask.im, mask.step))

    def get_32f_copy(self,Size size):
        cdef FastImage32f result
        result = FastImage32f(size)
        CHK_HAVEGIL( ipp.ippiConvert_8u32f_C1R(<ipp.Ipp8u*>self.im, self.step,
                                               <ipp.Ipp32f*>result.im, result.step,
                                               size.sz ))
        return result

    def get_32f_copy_put(self,FastImage32f result,Size size):
        CHK_HAVEGIL( ipp.ippiConvert_8u32f_C1R(<ipp.Ipp8u*>self.im, self.step,
                                               <ipp.Ipp32f*>result.im, result.step,
                                               size.sz ))

    cdef void fast_get_32f_copy_put(self,FastImage32f result,Size size):
        CHK_HAVEGIL( ipp.ippiConvert_8u32f_C1R(<ipp.Ipp8u*>self.im, self.step,
                                             <ipp.Ipp32f*>result.im, result.step,
                                             size.sz ))

    def get_8u_copy(self,Size size):
        cdef FastImage8u result
        result = FastImage8u(size)
        CHK_HAVEGIL( ipp.ippiCopy_8u_C1R(<ipp.Ipp8u*>self.im, self.step,
                                         <ipp.Ipp8u*>result.im, result.step,
                                         size.sz ))
        return result

    def get_8u_copy_put(self,FastImage8u result,Size size):
        CHK_HAVEGIL( ipp.ippiCopy_8u_C1R(<ipp.Ipp8u*>self.im, self.step,
                                         <ipp.Ipp8u*>result.im, result.step,
                                         size.sz ))

    def get_absdiff_put(self,FastImage8u other,FastImage8u result,Size size):
        """result = |self - other|"""
        CHK_HAVEGIL( ipp.ippiAbsDiff_8u_C1R(<ipp.Ipp8u*>other.im, other.step,
                                            <ipp.Ipp8u*>self.im, self.step,
                                            <ipp.Ipp8u*>result.im, result.step,
                                            size.sz))

    cdef void fast_get_absdiff_put(self, FastImage8u other, FastImage8u result, Size size):
        CHK_HAVEGIL( ipp.ippiAbsDiff_8u_C1R(<ipp.Ipp8u*>other.im,other.step,
                                          <ipp.Ipp8u*>self.im,self.step,
                                          <ipp.Ipp8u*>result.im,result.step,
                                          size.sz))

    def get_sub_put(self, FastImage8u other, FastImage8u result, Size size):
        """result = self - other"""
        CHK_HAVEGIL( ipp.ippiSub_8u_C1RSfs(<ipp.Ipp8u*>other.im, other.step,
                                           <ipp.Ipp8u*>self.im, self.step,
                                           <ipp.Ipp8u*>result.im, result.step,
                                           size.sz,0))
    cdef void fast_get_sub_put(self, FastImage8u other, FastImage8u result, Size size):
        CHK_HAVEGIL( ipp.ippiSub_8u_C1RSfs(<ipp.Ipp8u*>other.im, other.step,
                                         <ipp.Ipp8u*>self.im, self.step,
                                         <ipp.Ipp8u*>result.im, result.step,
                                         size.sz,0))

    def toself_threshold( self, int compare_value, int new_value, Size size, int op):
        """self = self (op) compare_value ? new_value : self"""
        print 'compare_value',compare_value
        print 'new_value',new_value
        CHK_HAVEGIL( ipp.ippiThreshold_Val_8u_C1IR(<ipp.Ipp8u*>self.im,self.step,
                                                   size.sz, compare_value, new_value,
                                                   <ipp.IppCmpOp>op))

    def get_compare( self, other, Size size, int op ):
        cdef FastImage8u result
        result = FastImage8u(size)
        self.get_compare_put(other, result, size, op)
        return result

    def get_compare_put( self, other, FastImage8u dest, Size size, int op ):
        cdef FastImage8u other8u
        cdef int other_int
        if type(other)==int:
            other_int = other
            CHK_HAVEGIL( ipp.ippiCompareC_8u_C1R( <ipp.Ipp8u*>self.im,self.step,
                                                  other_int,
                                                  <ipp.Ipp8u*>dest.im,dest.step,
                                                  size.sz,<ipp.IppCmpOp>op))
        elif type(other)==FastImage8u:
            other8u = other
            CHK_HAVEGIL( ipp.ippiCompare_8u_C1R( <ipp.Ipp8u*>self.im,self.step,
                                                 <ipp.Ipp8u*>other8u.im,other8u.step,
                                                 <ipp.Ipp8u*>dest.im,dest.step,
                                                 size.sz,<ipp.IppCmpOp>op))
        else:
            raise TypeError("can only compare against ints and FastImage8u types")

    cdef void fast_get_compare_int_put_greater( self, int other_int, FastImage8u dest, Size size):
        CHK_HAVEGIL( ipp.ippiCompareC_8u_C1R( <ipp.Ipp8u*>self.im,self.step,
                                            other_int,
                                            <ipp.Ipp8u*>dest.im,dest.step,
                                            size.sz,
                                            ipp.ippCmpGreater))

    def min_index(self, Size size):
        cdef ipp.IppStatus sts
        cdef int index_x, index_y
        cdef ipp.Ipp8u min_val

        with nogil:
            sts = ipp.ippiMinIndx_8u_C1R(
                <ipp.Ipp8u*>self.im,self.step,
                 size.sz, &min_val, &index_x, &index_y)
        CHK_HAVEGIL(sts)
        return min_val, index_x, index_y

    def max_index(self, Size size):
        cdef fic.FicStatus sts
        cdef int index_x, index_y
        cdef ipp.Ipp8u max_val
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        with nogil:
            sts = fic.ficiMaxIndx_8u_C1R(
                <fic.Fic8u*>self.im,self.step,
                 sz, &max_val, &index_x, &index_y)
        CHK_FIC_HAVEGIL(sts)
        return max_val, index_x, index_y

    def to_yuv422(self):
        cdef FastImage8u y, u, v

        y = self.get_8u_copy(self.imsiz)
        u = asfastimage(128*numpy.ones_like(y))
        v = u
        return y,u,v

    def dot(self,FastImage8u other,Size size):
        cdef fic.Fic64f result
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        CHK_FIC_HAVEGIL( fic.ficiDotProd_8u64f_C1R(
            <fic.Fic8u*>self.im,self.step,
            <fic.Fic8u*>other.im,other.step,
            sz, &result ))
        return result

    def sobel_horiz(self, Size size, FastImage8u dest=None):
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        if dest is None:
            dest = FastImage8u( size )
        else:
            assert dest.size == size
        CHK_FIC_HAVEGIL( fic.ficiFilterSobelHoriz_8u_C1R(<fic.Fic8u*>self.im, self.step, <fic.Fic8u*>dest.im, dest.step, sz ))
        return dest

    def sobel_vert(self, Size size, FastImage8u dest=None):
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        if dest is None:
            dest = FastImage8u( size )
        else:
            assert dest.size == size
        CHK_FIC_HAVEGIL( fic.ficiFilterSobelVert_8u_C1R(<fic.Fic8u*>self.im, self.step, <fic.Fic8u*>dest.im, dest.step, sz ))
        return dest

cdef class FastImage32f(FastImageBase):

    def __cinit__(self,*args,**kw):
        self.strides[1] = 4
        self.basetype = '32f'

    def __init__(self, Size size ):
        global FASTIMAGEDEBUG

        FastImageBase.__init__(self, size)
        if FASTIMAGEDEBUG>=1:
            print self,'requesting 32f memory (size W:%d H:%d)'%(self.imsiz.sz.width,
                                                             self.imsiz.sz.height)
        self.im=ipp.ippiMalloc_32f_C1( size.sz.width, size.sz.height, &self.step )
        if self.im==NULL: raise MemoryError("Error allocating memory")
        self.strides[0] = self.step
        if FASTIMAGEDEBUG>=1:
            self.show_mem_and_step()
        self.view = 0

    def set_val(self, float val, Size size):
        CHK_HAVEGIL( ipp.ippiSet_32f_C1R( val, <ipp.Ipp32f*>self.im, self.step, size.sz ))

    def toself_add(self, FastImageBase other, Size size):
        """self += other"""
        if isinstance(other,FastImage8u):
            CHK_HAVEGIL( ipp.ippiAdd_8u32f_C1IR( <ipp.Ipp8u*>other.im, other.step,
                                                 <ipp.Ipp32f*>self.im, self.step,
                                                 size.sz))
        elif isinstance(other,FastImage32f):
            CHK_HAVEGIL( ipp.ippiAdd_32f_C1IR( <ipp.Ipp32f*>other.im, other.step,
                                               <ipp.Ipp32f*>self.im, self.step,
                                               size.sz))
        else:
            raise ValueError("type not supported")

    def toself_add_weighted(self, FastImageBase other, Size size, float alpha):
        """self = self*(1-alpha) + alpha*other"""
        if isinstance(other,FastImage8u):
            CHK_HAVEGIL( ipp.ippiAddWeighted_8u32f_C1IR( <ipp.Ipp8u*>other.im, other.step,
                                                         <ipp.Ipp32f*>self.im, self.step,
                                                         size.sz, alpha))
        elif isinstance(other,FastImage32f):
            CHK_HAVEGIL( ipp.ippiAddWeighted_32f_C1IR( <ipp.Ipp32f*>other.im, other.step,
                                                       <ipp.Ipp32f*>self.im, self.step,
                                                       size.sz, alpha))
        else:
            raise ValueError("type not supported")

    cdef void fast_toself_add_weighted_8u( self, FastImage8u other, Size size, float alpha):
        CHK_HAVEGIL( ipp.ippiAddWeighted_8u32f_C1IR( <ipp.Ipp8u*>other.im, other.step,
                                                   <ipp.Ipp32f*>self.im, self.step,
                                                   size.sz, alpha))

    cdef void fast_toself_add_weighted_32f(self, FastImage32f other, Size size, float alpha):
        CHK_HAVEGIL( ipp.ippiAddWeighted_32f_C1IR( <ipp.Ipp32f*>other.im, other.step,
                                                   <ipp.Ipp32f*>self.im, self.step,
                                                   size.sz, alpha))


    def toself_add_square(self, FastImage8u other, Size size):
        """self += other**2"""
        CHK_HAVEGIL( ipp.ippiAddSquare_8u32f_C1IR( <ipp.Ipp8u*>other.im, other.step,
                                                   <ipp.Ipp32f*>self.im, self.step,
                                                     size.sz))
    def get_square_put(self, FastImage32f result, Size size):
        """result = self**2"""
        CHK_HAVEGIL( ipp.ippiSqr_32f_C1R( <ipp.Ipp32f*>self.im, self.step,
                                          <ipp.Ipp32f*>result.im, result.step,
                                          size.sz))

    cdef void fast_get_square_put(self, FastImage32f result, Size size):
        CHK_HAVEGIL( ipp.ippiSqr_32f_C1R( <ipp.Ipp32f*>self.im, self.step,
                                        <ipp.Ipp32f*>result.im, result.step,
                                        size.sz))

    def get_square(self, Size size):
        """return self**2"""
        cdef FastImage32f result
        result = FastImage32f( size )
        CHK_HAVEGIL( ipp.ippiSqr_32f_C1R( <ipp.Ipp32f*>self.im, self.step,
                                          <ipp.Ipp32f*>result.im, result.step,
                                          size.sz))
        return result

    def toself_square(self,Size size):
        """self = self**2"""
        CHK_HAVEGIL( ipp.ippiSqr_32f_C1IR( <ipp.Ipp32f*>self.im, self.step,
                                          size.sz))
    cdef void fast_toself_square(self,Size size):
        CHK_HAVEGIL( ipp.ippiSqr_32f_C1IR( <ipp.Ipp32f*>self.im, self.step,
                                         size.sz))
    def toself_sqrt(self,Size size):
        """self = sqrt(self)"""
        CHK_HAVEGIL( ipp.ippiSqrt_32f_C1IR( <ipp.Ipp32f*>self.im, self.step,
                                            size.sz))

    def get_subtracted(self,FastImage32f other,Size size):
        """result = self - other"""
        cdef FastImage32f result
        result = FastImage32f( size )
        CHK_HAVEGIL( ipp.ippiSub_32f_C1R(<ipp.Ipp32f*>other.im, other.step,
                                          <ipp.Ipp32f*>self.im, self.step,
                                          <ipp.Ipp32f*>result.im, result.step,
                                          size.sz))
        return result

    def get_subtracted_put(self,FastImage32f other,FastImage32f result,Size size):
        """result = self - other"""
        CHK_HAVEGIL( ipp.ippiSub_32f_C1R(<ipp.Ipp32f*>other.im, other.step,
                                          <ipp.Ipp32f*>self.im, self.step,
                                          <ipp.Ipp32f*>result.im, result.step,
                                          size.sz))

    cdef void fast_get_subtracted_put(self,FastImage32f other,FastImage32f result,Size size):
        """result = self - other"""
        CHK_HAVEGIL( ipp.ippiSub_32f_C1R(<ipp.Ipp32f*>other.im, other.step,
                                          <ipp.Ipp32f*>self.im, self.step,
                                          <ipp.Ipp32f*>result.im, result.step,
                                          size.sz))

    def toself_subtract(self,FastImage32f other,Size size):
        """self = self - other"""
        CHK_HAVEGIL( ipp.ippiSub_32f_C1IR(<ipp.Ipp32f*>other.im, other.step,
                                          <ipp.Ipp32f*>self.im, self.step,
                                          size.sz))

    def get_8u_copy(self,Size size):
        """other = self"""
        cdef FastImage8u other
        other = FastImage8u(size)
        CHK_HAVEGIL( ipp.ippiConvert_32f8u_C1R(
            <ipp.Ipp32f*>self.im,self.step,
            <ipp.Ipp8u*>other.im,other.step,
            size.sz, ipp.ippRndNear ))
        return other

    def get_8u_copy_put(self,FastImage8u other,Size size):
        """other = self"""
        CHK_HAVEGIL( ipp.ippiConvert_32f8u_C1R(
            <ipp.Ipp32f*>self.im,self.step,
            <ipp.Ipp8u*>other.im,other.step,
            size.sz, ipp.ippRndNear ))

    cdef void fast_get_8u_copy_put(self,FastImage8u other,Size size):
        CHK_HAVEGIL( ipp.ippiConvert_32f8u_C1R(
            <ipp.Ipp32f*>self.im,self.step,
            <ipp.Ipp8u*>other.im,other.step,
            size.sz, ipp.ippRndNear ))

    def toself_multiply(self, float val, Size size):
        "self = self * C"
        CHK_HAVEGIL( ipp.ippiMulC_32f_C1IR(val, <ipp.Ipp32f*>self.im, self.step, size.sz))

    cdef void fast_toself_multiply(self, float val, Size size):
        CHK_HAVEGIL( ipp.ippiMulC_32f_C1IR(val, <ipp.Ipp32f*>self.im, self.step, size.sz))

    def get_multiply_put(self, float val, FastImage32f other, Size size):
        "other = self * C"
        CHK_HAVEGIL( ipp.ippiMulC_32f_C1R(<ipp.Ipp32f*>self.im, self.step,  val,
                                          <ipp.Ipp32f*>other.im,other.step, size.sz))

    def pow(self,y):
        cdef FastImage32f xself
        xself = copy(self)
        if y == 2:
            xself.toself_square(xself.imsiz)
        elif y==0.5:
            xself.toself_sqrt(xself.imsiz)
        else:
            raise ValueError("only 2 and 0.5 powers supported")
        return xself

    def max_index(self, Size size):
        cdef fic.FicStatus sts
        cdef int index_x, index_y
        cdef ipp.Ipp32f max_val
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        with nogil:
            sts = fic.ficiMaxIndx_32f_C1R(
                <fic.Fic32f*>self.im,self.step,
                 sz, &max_val, &index_x, &index_y)
        CHK_FIC_HAVEGIL(sts)
        return max_val, index_x, index_y

    def dot(self,FastImage32f other,Size size):
        cdef fic.Fic64f result
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        CHK_FIC_HAVEGIL( fic.ficiDotProd_32f64f_C1R(
            <fic.Fic32f*>self.im,self.step,
            <fic.Fic32f*>other.im,other.step,
            sz, &result ))
        return result

    cdef FastImage32f create_equal_shape_empty(self, Size size):
        cdef FastImage32f dest
        dest = FastImage32f( size )
        if dest.step != self.step:
            shape = ( size.sz.height, self.step/4 )
            dtype = numpy.float32
            tmp = numpy.empty( shape, dtype=dtype)
            tmp = tmp[:,:size.sz.width]
            dest = asfastimage(tmp)
        return dest

    def sobel_horiz(self, Size size, FastImage32f dest=None):
        cdef fic.FiciSize sz
        cdef int created_dest

        created_dest = 0

        sz.width = size.sz.width
        sz.height = size.sz.height
        if dest is None:
            dest = self.create_equal_shape_empty(size)
        else:
            assert dest.size == size
        CHK_FIC_HAVEGIL( fic.ficiFilterSobelHoriz_32f_C1R(<fic.Fic32f*>self.im, self.step, <fic.Fic32f*>dest.im, dest.step, sz ))
        return dest

    def sobel_vert(self, Size size, FastImage32f dest=None):
        cdef fic.FiciSize sz

        sz.width = size.sz.width
        sz.height = size.sz.height
        if dest is None:
            dest = self.create_equal_shape_empty(size)
        else:
            assert dest.size == size
        CHK_FIC_HAVEGIL( fic.ficiFilterSobelVert_32f_C1R(<fic.Fic32f*>self.im, self.step, <fic.Fic32f*>dest.im, dest.step, sz ))
        return dest

def asfastimage( object arr ):
    """return a FastImage view of an array"""
    cdef PyArrayInterface* inter
    cdef FastImageBase result

    global FASTIMAGEDEBUG

    attr = arr.__array_struct__
    if not c_python.PyCObject_Check(attr):
        raise ValueError("invalid __array_struct__")

    inter = <PyArrayInterface*>c_python.PyCObject_AsVoidPtr(attr)
    if inter.two != 2:
        raise ValueError("invalid __array_struct__")

    if (inter.flags & ARR_HAS_DESCR) == (ARR_HAS_DESCR):
        raise ValueError("cannot handle arrays with descr")

    # TODO: don't really know what these flags all mean, figure out if
    # this is OK:
    if (inter.flags & (ALIGNED | WRITEABLE)) != (ALIGNED | WRITEABLE):
        raise ValueError("cannot handle misaligned or not writeable arrays.")

    if inter.nd != 2:
        raise ValueError("only 2D arrays may be viewed as a FastImage")

    if FASTIMAGEDEBUG>=2:
        print 'asfastimage() calling __new__'
    if inter.typekind == CHAR_u and inter.itemsize==1:
        result = FastImage8u.__new__(FastImage8u)
##    elif inter.typekind == "b"[0] and inter.itemsize==1: # "bool" true or false
##        result = FastImage8u.__new__(FastImage8u)
    elif inter.typekind == CHAR_f and inter.itemsize==4:
        result = FastImage32f.__new__(FastImage32f)
    else:
        raise ValueError('typekind "%s", itemsize %d not supported'%(chr(inter.typekind),inter.itemsize))
    if (inter.shape[0] > INT_MAX) or (inter.shape[1] > INT_MAX) or (inter.strides[0] > INT_MAX):
        raise ValueError("cannot handle such large data")
    FastImageBase.__init__(result, Size(inter.shape[1],inter.shape[0])) # run common __init__
    result.strides[0] = inter.strides[0]
    result.step = inter.strides[0]
    result.im = inter.data

    result.source_data = arr # keep reference to original image in case it goes out of scope elsewhere

    if FASTIMAGEDEBUG>=2:
        print '  result',result
        print '    result.view',result.view

    return result

def copy( object arr ):
    """return a FastImage copy of an array"""
    cdef PyArrayInterface* inter
    cdef FastImageBase result
    cdef int i

    attr = arr.__array_struct__
    if not c_python.PyCObject_Check(attr):
        raise ValueError("invalid __array_struct__")

    inter = <PyArrayInterface*>c_python.PyCObject_AsVoidPtr(attr)
    if inter.two != 2:
        raise ValueError("invalid __array_struct__")

    if (inter.flags & ARR_HAS_DESCR) == (ARR_HAS_DESCR):
        raise ValueError("cannot handle arrays with descr")

    # TODO: don't really know what these flags all mean, figure out if
    # this is OK:
    if (inter.flags & (ALIGNED)) != (ALIGNED):
        raise ValueError("cannot handle misaligned arrays.")

    if inter.nd != 2:
        raise ValueError("only 2D arrays may be copied to a FastImage")

    if (inter.shape[0] > INT_MAX) or (inter.shape[1] > INT_MAX) or (inter.strides[0] > INT_MAX):
        raise ValueError("cannot handle such large data")

    if inter.typekind == CHAR_u and inter.itemsize==1:
        result = FastImage8u(Size(inter.shape[1],inter.shape[0])) # allocate memory
    elif inter.typekind == CHAR_f and inter.itemsize==4:
        result = FastImage32f(Size(inter.shape[1],inter.shape[0])) # allocate memory
    else:
        raise ValueError('typekind "%s", itemsize %d not supported'%(chr(inter.typekind),inter.itemsize))

    # TODO: make sure that we can rely on blocks like this:
    for i from 0 <= i < inter.shape[0]:
        c_lib.memcpy( result.im + i*result.step,
                      inter.data + i*inter.strides[0],
                      inter.shape[1]*result.strides[1] )

    return result



import pkg_resources
import unittest
import motmot.FastImage.FastImage as fi
import numpy as np

class TestIPP(unittest.TestCase):
    def test_version(self):
        major,minor,build = fi.get_IPP_version()
    def test_arch(self):
        arch = fi.get_IPP_arch()
    def test_static(self):
        result = fi.get_IPP_static()

class TestSize(unittest.TestCase):
    def test_size_eq(self):
        s1 = fi.Size(432,42)
        s2 = fi.Size(432,42)
        s3 = fi.Size(432,43)
        self.assert_(s1 == s1)
        self.assert_(s1 is s1)
        self.assert_(s1 == s2)
        self.assert_(s1 is not s2)
        self.assert_(s1 != s3)

class TestFastImageBase(unittest.TestCase):
    def setUp(self):
        self.fastimagefactories = [fi.FastImage8u,
                                   fi.FastImage32f]
        self.ar_dtypes = [np.uint8,
                          np.float32]

    def test_roi(self):
        sz = fi.Size(20,10)
        for fif in self.fastimagefactories:
            imA=fif(sz)
            imA.set_val(87,sz)

            bottom = 3
            left = 14
            onepix = fi.Size(1,1)
            imB=imA.roi(left,bottom,onepix)
            imB.set_val(12,onepix)

            arA = 87*np.ones((sz.h, sz.w),np.uint8)
            arA[bottom,left] = 12
            self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_roi2(self):
        sz = fi.Size(20,10)
        for fif in self.fastimagefactories:
            imA=fif(sz)
            imA.set_val(87,sz)

            bottom = 3
            left = 14
            onepix = fi.Size(1,1)
            imB=imA.roi(left,bottom,onepix)
            del imA
            imB.set_val(12,onepix)

            arA = 87*np.ones((sz.h, sz.w),np.uint8)
            arB = arA[bottom,left]
            del arA
            arB = 12
            self.assert_( np.allclose(arB, np.asarray(imB)))

    def test_basic(self):
        size = fi.Size(20,10)
        for (fif,ar_dtype) in zip(self.fastimagefactories,self.ar_dtypes):
            imA=fif(size)
            imA.set_val(0,size)
            arA = np.zeros((size.h, size.w),ar_dtype)
            self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_getitem(self):
        size = fi.Size(20,10)
        for fif in self.fastimagefactories:
            imA=fif(size)
            imA.set_val(0,size)
            self.assert_( 0==imA[0,0] )

    def test_mem(self):
        size = fi.Size(20,10)
        for (fif,ar_dtype) in zip(self.fastimagefactories,self.ar_dtypes):
            imA=fif(size)
            imA.set_val(0,size)
            arA = np.zeros((size.h, size.w),ar_dtype)

            imB=fi.copy(imA)
            imB.set_val(2,size)
            self.assert_( not np.allclose(np.asarray(imA),np.asarray(imB)) )

class TestFastImage8u(unittest.TestCase):

    def test_32f_copy(self):
        sz = fi.Size(20,10)

        imA=fi.FastImage8u(sz)
        imA.set_val(87,sz)

        imB=fi.FastImage32f(sz)
        imA.get_32f_copy_put(imB,sz)

        arB = 87*np.ones((sz.h, sz.w),np.float)
        self.assert_( np.allclose(arB, np.asarray(imB)))

    def test_dot(self):

        A = 254
        B = 255
        sz = fi.Size(33,323)

        imA=fi.FastImage8u(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage8u(sz)
        imB.set_val(B,sz)

        C = imA.dot(imB,sz)

        # Do comparison in floating point to ensure no integer wrap-
        # around bugs.

        arA = A*np.ones((sz.h, sz.w),np.float)
        arB = B*np.ones((sz.h, sz.w),np.float)
        arC = np.sum(arA.ravel() * arB.ravel())
        self.assert_( abs(C-arC) < 1e-16 )

    def test_absdiff(self):

        A = 32
        B = 38
        sz = fi.Size(33,323)

        imA=fi.FastImage8u(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage8u(sz)
        imB.set_val(B,sz)

        imC=fi.FastImage8u(sz)
        imA.get_absdiff_put(imB,imC,sz)

        arA = A*np.ones((sz.h, sz.w),np.int16)
        arB = B*np.ones((sz.h, sz.w),np.int16)
        arC = abs(arA - arB)
        self.assert_( np.allclose(arC, np.asarray(imC)))

    def test_absdiff2(self):

        A = 32
        B = 38
        sz = fi.Size(656,491)

        imA=fi.FastImage8u(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage8u(sz)
        imB.set_val(B,sz)

        imC=fi.FastImage8u(sz)
        imA.get_absdiff_put(imB,imC,sz)

        arA = A*np.ones((sz.h, sz.w),np.int16)
        arB = B*np.ones((sz.h, sz.w),np.int16)
        arC = abs(arA - arB)
        self.assert_( np.allclose(arC, np.asarray(imC)))

    def test_sub(self):

        A = 32
        B = 38
        sz = fi.Size(656,491)

        imA=fi.FastImage8u(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage8u(sz)
        imB.set_val(B,sz)

        imC=fi.FastImage8u(sz)
        imA.get_sub_put(imB,imC,sz) # imC = imA - imB

        imD=fi.FastImage8u(sz)
        imB.get_sub_put(imA,imD,sz) # imD = imB - im A

        # perform signed int math then clip to emulate
        arA = A*np.ones((sz.h, sz.w),np.int16)
        arB = B*np.ones((sz.h, sz.w),np.int16)
        arC = np.clip(arA - arB,0,255)
        arD = np.clip(arB - arA,0,255)
        self.assert_( np.allclose(arC, np.asarray(imC)))
        self.assert_( np.allclose(arD, np.asarray(imD)))

    def test_min_index(self):
        sz = fi.Size(5,5)
        L = fi.FastImage8u(sz)
        imL = np.asarray(L)
        imL[:,:] = 100
        imL[2,4] = 10
        val,x,y=L.min_index(sz)
        self.assert_(val==10)
        self.assert_(x==4)
        self.assert_(y==2)

    def test_max_index(self):
        sz = fi.Size(5,5)
        L = fi.FastImage8u(sz)
        imL = np.asarray(L)
        imL[:,:] = 10
        imL[2,4] = 100
        val,x,y=L.max_index(sz)
        self.assert_(val==100)
        self.assert_(x==4)
        self.assert_(y==2)

    def test_mono8_to_yuv422(self):
        sz = fi.Size(5,5)

        L = fi.FastImage8u(sz)
        L.set_val(100,sz)

        y,u,v = L.to_yuv422()

class TestFastImage32f(unittest.TestCase):
    def test_roi(self):
        sz = fi.Size(20,10)

        imA=fi.FastImage32f(sz)
        imA.set_val(87,sz)

        bottom = 3
        left = 14
        onepix = fi.Size(1,1)
        imB=imA.roi(left,bottom,onepix)
        imB.set_val(12,onepix)

        arA = 87*np.ones((sz.h, sz.w),np.float32)
        arA[bottom,left]=12
        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_dot(self):

        A = 254
        B = 255
        sz = fi.Size(33,323)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage32f(sz)
        imB.set_val(B,sz)

        C = imA.dot(imB,sz)

        # Do comparison in floating point to ensure no integer wrap-
        # around bugs.

        arA = A*np.ones((sz.h, sz.w),np.float)
        arB = B*np.ones((sz.h, sz.w),np.float)
        arC = np.sum(arA.ravel() * arB.ravel())
        self.assert_( abs(C-arC) < 1e-16 )

    def test_add_weighted_8u(self):

        A = 32.2
        B = 83
        alpha = 0.3
        sz = fi.Size(20,10)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage8u(sz)
        imB.set_val(B,sz)

        imA.toself_add_weighted(imB,sz,alpha)

        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.uint8)
        arA = arA*(1.0-alpha) + arB*alpha
        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_add_weighted_32f(self):

        A = 32.2
        B = 3854.2
        alpha = 0.3
        sz = fi.Size(33,33)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage32f(sz)
        imB.set_val(B,sz)

        imA.toself_add_weighted(imB,sz,alpha)

        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.float32)
        arA = arA*(1.0-alpha) + arB*alpha
        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_squares(self):
        A = 32.2
        sz = fi.Size(5,5)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)
        imA.toself_square(sz)
        imB=imA.get_square(sz)
        imC=fi.FastImage32f(sz)
        imB.get_square_put(imC,sz)

        arA=A*np.ones((sz.h, sz.w),np.float32)
        arC=(arA**8).astype(np.float32)
        self.assert_( np.allclose(arC, np.asarray(imC)))

    def test_sqrt(self):
        A = 32.2
        sz = fi.Size(5,245)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)
        imA.toself_sqrt(sz)

        arA=A*np.ones((sz.h, sz.w),np.float32)
        arA=(np.sqrt(arA)).astype(np.float32)
        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_add_8u(self):

        A = 32.2
        B = 82
        sz = fi.Size(63,33)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage8u(sz)
        imB.set_val(B,sz)

        imA.toself_add(imB,sz)
        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.uint8)
        arA = arA + arB
        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_add_32f(self):

        A = 32.2
        B = 3854.2
        sz = fi.Size(33,63)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage32f(sz)
        imB.set_val(B,sz)

        imA.toself_add(imB,sz)
        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.float32)
        arA = arA + arB
        self.assert_( np.allclose(arA, np.asarray(imA)))


    def test_subtract1(self):

        A = 32.2
        B = 3854.2
        sz = fi.Size(33,323)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage32f(sz)
        imB.set_val(B,sz)

        imA.toself_subtract(imB,sz)

        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.float32)
        arA = arA - arB
        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_subtract2(self):

        A = 32.2
        B = 3854.2
        sz = fi.Size(33,323)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage32f(sz)
        imB.set_val(B,sz)

        imC=imA.get_subtracted(imB,sz)

        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.float32)
        arC = arA - arB
        self.assert_( np.allclose(arC, np.asarray(imC)))

    def test_subtract3(self):

        A = 32.2
        B = 3854.2
        sz = fi.Size(33,323)

        imA=fi.FastImage32f(sz)
        imA.set_val(A,sz)

        imB=fi.FastImage32f(sz)
        imB.set_val(B,sz)

        imC=fi.FastImage32f(sz)
        imA.get_subtracted_put(imB,imC,sz)

        arA = A*np.ones((sz.h, sz.w),np.float32)
        arB = B*np.ones((sz.h, sz.w),np.float32)
        arC = arA - arB
        self.assert_( np.allclose(arC, np.asarray(imC)))

    def test_from_nx1(self):

        A = np.array(32.2,np.float32)
        sz = fi.Size(33,323)
        arA = np.array(A,np.float32)*np.ones((sz.h, sz.w),np.float32)

        imA=fi.asfastimage(arA)
        arA[1,3:40] = 3024.03

        self.assert_( np.allclose(arA, np.asarray(imA)))

    def test_from_nx2(self):

        A = np.array(32.2,np.float32)
        sz = fi.Size(33,323)
        arA = np.array(A,np.float32)*np.ones((sz.h, sz.w),np.float32)

        imA=fi.copy(arA)
        arA[1,3:40] = 3024.03

        self.assert_( not np.allclose(arA, np.asarray(imA)))

    def test_pow(self):
        sz = fi.Size(2,2)
        imB=fi.FastImage32f(sz)
        imB.set_val(2,sz)
        arB = np.array(imB)
        imC = imB.pow(2)
        arC = arB**2

        self.assert_( np.allclose(arB, np.asarray(imB)))
        self.assert_( np.allclose(arC, np.asarray(imC)))

    ## def test_min_index(self):
    ##     sz = fi.Size(5,5)
    ##     L = fi.FastImage32f(sz)
    ##     imL = np.asarray(L)
    ##     imL[:,:] = 100
    ##     imL[2,4] = 10
    ##     val,x,y=L.min_index(sz)
    ##     self.assert_(val==10)
    ##     self.assert_(x==4)
    ##     self.assert_(y==2)

    def test_max_index(self):
        sz = fi.Size(5,5)
        L = fi.FastImage32f(sz)
        imL = np.asarray(L)
        imL[:,:] = 10
        imL[2,4] = 100
        val,x,y=L.max_index(sz)
        self.assert_(val==100)
        self.assert_(x==4)
        self.assert_(y==2)

def test_get_8u_copy():
    a = np.linspace(0,255,40000).astype(np.float32)
    a.shape=200,200

    imA=fi.asfastimage(a)
    imB=imA.get_8u_copy(imA.size)
    b_test = np.asarray(imB)
    diff = a-b_test
    assert np.all(abs(diff)<1.0)

def test_conversion_step():
    sz = fi.Size(660,480)
    cls_dtype_nbytes = [ (fi.FastImage8u, np.uint8, 1),
                         (fi.FastImage32f, np.float32, 4),
                         ]
    for cls, dtype, nbytes in cls_dtype_nbytes:
        imA=cls(sz)

        npB = np.empty( (sz.h,imA._step//nbytes), dtype=dtype )
        npB = npB[:,:sz.w]

        imB = fi.asfastimage( npB )

        assert imA._step == imB._step
        assert imA.size == imB.size

def test_sobel_vert_1():
    a = np.linspace(0,255,40000).astype(np.float32)
    a.shape=200,200
    imA=fi.asfastimage(a)
    result = np.asarray(imA.sobel_vert(imA.size))

    b = np.ones_like(a)
    imB=fi.asfastimage(b)
    imA.sobel_vert(imA.size,dest=imB)
    assert np.allclose(result,b)

def test_sobel_horiz_1():
    a = np.linspace(0,255,40000).astype(np.float32)
    a.shape=200,200
    imA=fi.asfastimage(a)
    result = np.asarray(imA.sobel_horiz(imA.size))

    b = np.ones_like(a)
    imB=fi.asfastimage(b)
    imA.sobel_horiz(imA.size,dest=imB)
    assert np.allclose(result,b)

def get_test_suite():
    ts=unittest.TestSuite([
            unittest.makeSuite(TestIPP),
            unittest.makeSuite(TestSize),
            unittest.makeSuite(TestFastImageBase),
            unittest.makeSuite(TestFastImage8u),
            unittest.makeSuite(TestFastImage32f),
            ])
    return ts

if __name__ == '__main__':
    if 1:
        ts = get_test_suite()
        ts.debug()
    else:
        unittest.main()



import glob, os, sys

def get_build_info(ipp_static=True,  # static build requires static IPP libs
                   ipp_arch=None,
                   ipp_root=None,
                   ):
    """get options to build Python extensions built with Intel IPP

ipp_static - True to build using static IPP library (requires IPP license)
ipp_arch - Architecture of IPP to use (None uses default, 'intel64' and 'ia32' are options)
ipp_root - Root location of the IPP installation (e.g. /opt/intel/compilers_and_libraries/linux/ipp)
"""

    if ipp_arch is None:
        if sys.platform == 'darwin':
            ipp_arch = 'intel64'
        elif sys.platform.startswith('linux'):
            machine = os.uname()[4]
            if machine == 'x86_64':
                ipp_arch = 'intel64'
            elif machine in ['i386','i686']:
                ipp_arch = 'ia32'
            else:
                raise ValueError("unexpected linux architecture: %s"%machine)
        elif sys.platform == 'win32':
            ipp_arch = 'ia32'
        else:
            raise NotImplementedError("auto-architecture detection not implemented on this platform")

    vals = {}

    if sys.platform.startswith('linux'):
        libdirname = 'lib/%s_lin' % ipp_arch
    else:
        libdirname = 'lib'
    incdirname = 'include'
    ipp_define_macros = []
    ipp_extra_link_args = []
    ipp_extra_objects = []
    if sys.platform.startswith('win'):
        ipp_define_macros = [('FASTIMAGE_IPP_ARCH','\\"%s\\"'%ipp_arch)]
    else:
        ipp_define_macros = [('FASTIMAGE_IPP_ARCH','"%s"'%ipp_arch)]

    if ipp_static:
        ipp_define_macros.append( ('FASTIMAGE_STATICIPP','1') )
    ipp_include_dirs = [os.path.join(ipp_root,incdirname)]

    #  like LDFLAGS in sample Makefile.osx

    if ipp_static:
        ipp_library_dirs = []
        ipp_libraries = []
        libdir = os.path.join(ipp_root,libdirname)
        ipp_extra_objects = [os.path.join(libdir,'lib'+lib+'.a') for lib in 'ippi','ipps','ippcv','ippcc','ippcore']
    else:
        ipp_library_dirs = [ os.path.join(ipp_root,libdirname) ]
        ipp_libraries = ['ippi','ipps','ippcv','ippcc','ippcore']
        ipp_extra_objects = []

    vals['extra_link_args'] = ipp_extra_link_args
    vals['ipp_library_dirs'] = ipp_library_dirs
    vals['ipp_libraries'] = ipp_libraries
    vals['ipp_define_macros'] = ipp_define_macros
    vals['ipp_include_dirs'] = ipp_include_dirs
    vals['ipp_extra_objects'] = ipp_extra_objects
    return vals



import os, sys
from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize

# I need to import a module I'm about to install. How naughty is this?
sys.path.insert(0, os.path.abspath(os.path.join('motmot','FastImage')))
import util as FastImage_util

ipp_root = os.environ['IPPROOT']
vals = FastImage_util.get_build_info(ipp_static=True, ipp_root=ipp_root)

setup(name="motmot.FastImage",
      author="Andrew Straw",
      author_email="strawman@astraw.com",
      description="Pythonic API for a subset of the Intel "\
      "Integrated Performance Primitives (Intel IPP) library",
      url='http://code.astraw.com/projects/motmot',
      license="BSD",
      version='0.6.1', # also in motmot/FastImage/__init__.py
      namespace_packages=['motmot'],
      packages = find_packages(),
      ext_modules=cythonize([Extension(name="motmot.FastImage.FastImage",
                                       sources=['src/FastImage.pyx',
                                                'src/fic.c','src/fic_sobel.c'],
                                       include_dirs=vals['ipp_include_dirs'],
                                       library_dirs=vals['ipp_library_dirs'],
                                       libraries=vals['ipp_libraries'],
                                       define_macros=vals['ipp_define_macros'],
                                       extra_link_args=vals['extra_link_args'],
                                       extra_objects=vals['ipp_extra_objects'],
                                       ),
                             ]),
)



#emacs, this is -*-Python-*- mode

# WARNING   WARNING   WARNING
#
# If you add any functions to this file, also add them to funclist.h.
#
# WARNING   WARNING   WARNING

#cdef extern from "ippdefs.h":
cdef extern from "fi_ipp.h" nogil:
    ctypedef enum IppStatus:
        ippStsNotSupportedModeErr
        ippStsCpuNotSupportedErr

        ippStsEphemeralKeyErr
        ippStsMessageErr
        ippStsShareKeyErr
        ippStsIvalidPublicKey
        ippStsIvalidPrivateKey
        ippStsOutOfECErr
        ippStsECCInvalidFlagErr

        ippStsMP3FrameHeaderErr
        ippStsMP3SideInfoErr

        ippStsBlockStepErr
        ippStsMBStepErr

        ippStsAacPrgNumErr
        ippStsAacSectCbErr
        ippStsAacSfValErr
        ippStsAacCoefValErr
        ippStsAacMaxSfbErr
        ippStsAacPredSfbErr
        ippStsAacPlsDataErr
        ippStsAacGainCtrErr
        ippStsAacSectErr
        ippStsAacTnsNumFiltErr
        ippStsAacTnsLenErr
        ippStsAacTnsOrderErr
        ippStsAacTnsCoefResErr
        ippStsAacTnsCoefErr
        ippStsAacTnsDirectErr
        ippStsAacTnsProfileErr
        ippStsAacErr
        ippStsAacBitOffsetErr
        ippStsAacAdtsSyncWordErr
        ippStsAacSmplRateIdxErr
        ippStsAacWinLenErr
        ippStsAacWinGrpErr
        ippStsAacWinSeqErr
        ippStsAacComWinErr
        ippStsAacStereoMaskErr
        ippStsAacChanErr
        ippStsAacMonoStereoErr
        ippStsAacStereoLayerErr
        ippStsAacMonoLayerErr
        ippStsAacScalableErr
        ippStsAacObjTypeErr
        ippStsAacWinShapeErr
        ippStsAacPcmModeErr
        ippStsVLCUsrTblHeaderErr
        ippStsVLCUsrTblUnsupportedFmtErr
        ippStsVLCUsrTblEscAlgTypeErr
        ippStsVLCUsrTblEscCodeLengthErr
        ippStsVLCUsrTblCodeLengthErr
        ippStsVLCInternalTblErr
        ippStsVLCInputDataErr
        ippStsVLCAACEscCodeLengthErr
        ippStsNoiseRangeErr
        ippStsUnderRunErr
        ippStsPaddingErr
        ippStsCFBSizeErr
        ippStsPaddingSchemeErr
        ippStsInvalidCryptoKeyErr
        ippStsLengthErr
        ippStsBadModulusErr
        ippStsLPCCalcErr
        ippStsRCCalcErr
        ippStsIncorrectLSPErr
        ippStsNoRootFoundErr
        ippStsJPEG2KBadPassNumber
        ippStsJPEG2KDamagedCodeBlock
        ippStsH263CBPYCodeErr
        ippStsH263MCBPCInterCodeErr
        ippStsH263MCBPCIntraCodeErr
        ippStsNotEvenStepErr
        ippStsHistoNofLevelsErr
        ippStsLUTNofLevelsErr
        ippStsMP4BitOffsetErr
        ippStsMP4QPErr
        ippStsMP4BlockIdxErr
        ippStsMP4BlockTypeErr
        ippStsMP4MVCodeErr
        ippStsMP4VLCCodeErr
        ippStsMP4DCCodeErr
        ippStsMP4FcodeErr
        ippStsMP4AlignErr
        ippStsMP4TempDiffErr
        ippStsMP4BlockSizeErr
        ippStsMP4ZeroBABErr
        ippStsMP4PredDirErr
        ippStsMP4BitsPerPixelErr
        ippStsMP4VideoCompModeErr
        ippStsMP4LinearModeErr
        ippStsH263PredModeErr
        ippStsH263BlockStepErr
        ippStsH263MBStepErr
        ippStsH263FrameWidthErr
        ippStsH263FrameHeightErr
        ippStsH263ExpandPelsErr
        ippStsH263PlaneStepErr
        ippStsH263QuantErr
        ippStsH263MVCodeErr
        ippStsH263VLCCodeErr
        ippStsH263DCCodeErr
        ippStsH263ZigzagLenErr
        ippStsFBankFreqErr
        ippStsFBankFlagErr
        ippStsFBankErr
        ippStsNegOccErr
        ippStsCdbkFlagErr
        ippStsSVDCnvgErr
        ippStsJPEGHuffTableErr
        ippStsJPEGDCTRangeErr
        ippStsJPEGOutOfBufErr
        ippStsDrawTextErr
        ippStsChannelOrderErr
        ippStsZeroMaskValuesErr
        ippStsQuadErr
        ippStsRectErr
        ippStsCoeffErr
        ippStsNoiseValErr
        ippStsDitherLevelsErr
        ippStsNumChannelsErr
        ippStsCOIErr
        ippStsDivisorErr
        ippStsAlphaTypeErr
        ippStsGammaRangeErr
        ippStsGrayCoefSumErr
        ippStsChannelErr
        ippStsToneMagnErr
        ippStsToneFreqErr
        ippStsTonePhaseErr
        ippStsTrnglMagnErr
        ippStsTrnglFreqErr
        ippStsTrnglPhaseErr
        ippStsTrnglAsymErr
        ippStsHugeWinErr
        ippStsJaehneErr
        ippStsStrideErr
        ippStsEpsValErr
        ippStsWtOffsetErr
        ippStsAnchorErr
        ippStsMaskSizeErr
        ippStsShiftErr
        ippStsSampleFactorErr
        ippStsSamplePhaseErr
        ippStsFIRMRFactorErr
        ippStsFIRMRPhaseErr
        ippStsRelFreqErr
        ippStsFIRLenErr
        ippStsIIROrderErr
        ippStsDlyLineIndexErr
        ippStsResizeFactorErr
        ippStsInterpolationErr
        ippStsMirrorFlipErr
        ippStsMoment00ZeroErr
        ippStsThreshNegLevelErr
        ippStsThresholdErr
        ippStsContextMatchErr
        ippStsFftFlagErr
        ippStsFftOrderErr
        ippStsStepErr
        ippStsScaleRangeErr
        ippStsDataTypeErr
        ippStsOutOfRangeErr
        ippStsDivByZeroErr
        ippStsMemAllocErr
        ippStsNullPtrErr
        ippStsRangeErr
        ippStsSizeErr
        ippStsBadArgErr
        ippStsNoMemErr
        ippStsSAReservedErr3
        ippStsErr
        ippStsSAReservedErr1

        ippStsNoErr


        ippStsNoOperation
        ippStsMisalignedBuf
        ippStsSqrtNegArg
        ippStsInvZero
        ippStsEvenMedianMaskSize
        ippStsDivByZero
        ippStsLnZeroArg
        ippStsLnNegArg
        ippStsNanArg
        ippStsJPEGMarker
        ippStsResFloor
        ippStsOverflow
        ippStsLSFLow
        ippStsLSFHigh
        ippStsLSFLowAndHigh
        ippStsZeroOcc
        ippStsUnderflow
        ippStsSingularity
        ippStsDomain
        ippStsNonIntelCpu
        ippStsCpuMismatch
        ippStsNoIppFunctionFound
        ippStsDllNotFoundBestUsed
        ippStsNoOperationInDll
        ippStsInsufficientEntropy
        ippStsOvermuchStrings
        ippStsOverlongString
        ippStsAffineQuadChanged

    ctypedef enum IppRoundMode:
        ippRndZero
        ippRndNear

    ctypedef enum IppiMaskSize:
        ippMskSize1x3
        ippMskSize1x5
        ippMskSize3x1
        ippMskSize3x3
        ippMskSize5x1
        ippMskSize5x5

    ctypedef enum IppiBayerGrid:
        ippiBayerBGGR
        ippiBayerRGGB
        ippiBayerGBRG
        ippiBayerGRBG

    ctypedef enum IppCpuType:
        ippCpuUnknown
        ippCpuPP
        ippCpuPMX
        ippCpuPPR
        ippCpuPII
        ippCpuPIII
        ippCpuP4
        ippCpuP4HT
        ippCpuP4HT2
        ippCpuCentrino
        ippCpuITP
        ippCpuITP2

    ctypedef enum IppCmpOp:
        ippCmpLess
        ippCmpLessEq
        ippCmpEq
        ippCmpGreaterEq
        ippCmpGreater

    ctypedef unsigned char   Ipp8u
    ctypedef unsigned short  Ipp16u
    ctypedef unsigned int    Ipp32u

    ctypedef signed char    Ipp8s
    ctypedef signed short   Ipp16s
    ctypedef signed int     Ipp32s
    ctypedef float   Ipp32f
#    ctypedef __INT64 Ipp64s
#    ctypedef __UINT64 Ipp64u
    ctypedef double  Ipp64f

    ctypedef struct IppiSize:
        int width
        int height

    ctypedef struct IppiRect:
        int x
        int y
        int width
        int height

    ctypedef struct IppiPoint:
        int x
        int y

    ctypedef enum IppHintAlgorithm:
        ippAlgHintNone
        ippAlgHintFast
        ippAlgHintAccurate

#cdef extern from "ippcore.h":
#    IppStatus ippStaticInit()
#    IppStatus ippStaticFree()
#    IppStatus ippStaticInitBest()
#    IppStatus ippStaticInitCpu( IppCpuType cpu )

#    IppStatus ippCoreSetFlushToZero( int value, unsigned int* pUMask )
#    IppStatus ippCoreSetDenormAreZeros( int value )

#    IppCpuType ippCoreGetCpuType()
    char* ippGetStatusString( IppStatus StsCode )

#cdef extern from "ippi.h":
    ctypedef struct IppiMomentState_64f
    ctypedef Ipp64f IppiHuMoment_64f[7]

    Ipp8u* ippiMalloc_8u_C1( int widthPixels, int heightPixels, int* pStepBytes )
    Ipp8u* ippiMalloc_8u_C3( int widthPixels, int heightPixels, int* pStepBytes )
    Ipp32f* ippiMalloc_32f_C1( int widthPixels, int heightPixels, int* pStepBytes )
    void ippiFree(void* ptr)
    IppStatus ippiSub_8u_C1IRSfs(Ipp8u* pSrc, int srcStep, Ipp8u* pSrcDst,
                                 int srcDstStep, IppiSize roiSize, int scaleFactor)
    IppStatus ippiSub_8u_C1RSfs( Ipp8u* pSrc1, int src1Step, Ipp8u* pSrc2,
                                 int src2Step, Ipp8u* pDst, int dstStep, IppiSize roiSize,
                                 int scaleFactor)
    IppStatus ippiSub_32f_C1IR(Ipp32f* pSrc, int srcStep, Ipp32f* pSrcDst,
                               int srcDstStep, IppiSize roiSize)
    IppStatus ippiSub_32f_C1R(Ipp32f* pSrc1, int src1Step, Ipp32f* pSrc2,
                              int src2Step, Ipp32f* pDst, int dstStep, IppiSize roiSize)

    IppStatus ippiAdd_8u_C1RSfs(Ipp8u* pSrc1, int src1Step, Ipp8u* pSrc2,
                                int src2Step, Ipp8u* pDst, int dstStep, IppiSize roiSize,
                                int scaleFactor)
    IppStatus ippiAdd_8u_C1IRSfs(Ipp8u* pSrc, int srcStep, Ipp8u* pSrcDst,
                                 int srcDstStep, IppiSize roiSize, int scaleFactor)
    IppStatus ippiAdd_32f_C1IR(Ipp32f* pSrc, int srcStep, Ipp32f* pSrcDst,
                               int srcDstStep, IppiSize roiSize)
    IppStatus ippiSet_8u_C1R( Ipp8u value, Ipp8u* pDst, int dstStep,
                              IppiSize roiSize )
    IppStatus ippiSet_8u_C1MR( Ipp8u value, Ipp8u* pDst, int dstStep,
                               IppiSize roiSize,
                               Ipp8u* pMask, int maskStep )
    IppStatus ippiSet_32f_C1R( Ipp32f value, Ipp32f* pDst, int dstStep,
                              IppiSize roiSize )
    IppStatus ippiConvert_8u32f_C1R(Ipp8u* pSrc, int srcStep, Ipp32f* pDst, int dstStep,
                                    IppiSize roiSize )
    IppStatus ippiConvert_32f8u_C1R(Ipp32f* pSrc, int srcStep, Ipp8u* pDst, int dstStep,
                                     IppiSize roiSize, IppRoundMode roundMode )
    IppStatus ippiThreshold_LT_8u_C1IR(Ipp8u* pSrcDst, int srcDstStep,
                                       IppiSize roiSize, Ipp8u threshold)
    IppStatus ippiThreshold_LT_32f_C1IR(Ipp32f* pSrcDst, int srcDstStep,
                                        IppiSize roiSize, Ipp32f threshold)
    IppStatus ippiThreshold_Val_8u_C1IR(Ipp8u* pSrcDst, int srcDstStep,
                                        IppiSize roiSize, Ipp8u threshold, Ipp8u value, IppCmpOp ippCmpOp)
    IppStatus ippiThreshold_Val_32f_C1IR(Ipp32f* pSrcDst, int srcDstStep,
                                         IppiSize roiSize, Ipp32f threshold, Ipp32f value, IppCmpOp ippCmpOp)
    IppStatus ippiMin_8u_C1R(Ipp8u* pSrc, int srcStep, IppiSize roiSize, Ipp8u* pMin)
    IppStatus ippiMin_32f_C1R(Ipp32f* pSrc, int srcStep, IppiSize roiSize, Ipp32f* pMin)
    IppStatus ippiMax_8u_C1R(Ipp8u* pSrc, int srcStep, IppiSize roiSize, Ipp8u* pMax)
    IppStatus ippiMax_32f_C1R(Ipp32f* pSrc, int srcStep, IppiSize roiSize, Ipp32f* pMax)
    IppStatus ippiCopy_8u_C1R(Ipp8u* pSrc, int srcStep,
                              Ipp8u* pDst, int dstStep, IppiSize roiSize )
    IppStatus ippiCopy_32f_C1R(Ipp32f* pSrc, int srcStep,
                               Ipp32f* pDst, int dstStep, IppiSize roiSize )
    IppStatus ippiMulC_32f_C1R(Ipp32f* pSrc, int srcStep, Ipp32f value, Ipp32f* pDst,
                               int dstStep, IppiSize roiSize)
    IppStatus ippiMulC_32f_C1IR(Ipp32f value, Ipp32f* pSrcDst, int srcDstStep,
                                IppiSize roiSize)
    IppStatus ippiCompareC_8u_C1R(Ipp8u* pSrc, int srcStep, Ipp8u value,
                                  Ipp8u* pDst, int dstStep,
                                  IppiSize roiSize,  IppCmpOp ippCmpOp)
    IppStatus ippiCompare_8u_C1R( Ipp8u* pSrc1, int src1Step,
                                  Ipp8u* pSrc2, int src2Step,
                                  Ipp8u* pDst,  int dstStep,
                                  IppiSize roiSize,   IppCmpOp ippCmpOp)
    IppStatus ippiCompare_32f_C1R(Ipp32f* pSrc1, int src1Step,
                                  Ipp32f* pSrc2, int src2Step,
                                  Ipp8u* pDst,  int dstStep,
                                  IppiSize roiSize,   IppCmpOp ippCmpOp)
    IppStatus ippiCompareEqualEps_32f_C1R( Ipp32f* pSrc1, int src1Step,
                                           Ipp32f* pSrc2, int src2Step,
                                           Ipp8u*  pDst,  int dstStep,
                                           IppiSize roiSize,    Ipp32f eps)
    IppStatus ippiDivC_8u_C1IRSfs(Ipp8u value, Ipp8u* pSrcDst,
                                  int srcDstStep, IppiSize roiSize, int scaleFactor)
    IppStatus ippiDivC_32f_C1R(Ipp32f* pSrc, int srcStep, Ipp32f value,
                               Ipp32f* pDst, int dstStep, IppiSize roiSize)
    IppStatus ippiMinIndx_8u_C1R(Ipp8u* pSrc, int srcStep, IppiSize roiSize, Ipp8u* pMin,
                                 int* pIndexX, int* pIndexY)
    IppStatus ippiMaxIndx_8u_C1R(Ipp8u* pSrc, int srcStep, IppiSize roiSize, Ipp8u* pMax,
                                 int* pIndexX, int* pIndexY)
    IppStatus ippiMaxIndx_32f_C1R(Ipp32f* pSrc, int srcStep, IppiSize roiSize,
                                  Ipp32f* pMax, int* pIndexX, int* pIndexY)
    IppStatus ippiSqr_32f_C1R(Ipp32f* pSrc, int srcStep,
                              Ipp32f* pDst, int dstStep, IppiSize roiSize)
    IppStatus ippiSqr_32f_C1IR(Ipp32f* pSrcDst, int srcDstStep,
                              IppiSize roiSize)
    IppStatus ippiSqr_8u_C1IRSfs(Ipp8u* pSrcDst, int srcDstStep,
                              IppiSize roiSize, int scaleFactor)
    IppStatus ippiSqrt_32f_C1IR(Ipp32f* pSrcDst, int srcDstStep,
                                IppiSize roiSize)
    IppStatus ippiMomentGetStateSize_64f(IppHintAlgorithm hint,int* stateSize)
    void* ippMalloc(int size)
    void ippFree(void* ptr)
    IppStatus ippiMomentInit_64f(IppiMomentState_64f* pState, IppHintAlgorithm hint)
    IppStatus ippiMoments64f_8u_C1R( Ipp8u*  pSrc, int srcStep, IppiSize roiSize, IppiMomentState_64f* pCtx)
    IppStatus ippiMoments64f_32f_C1R(Ipp32f* pSrc, int srcStep, IppiSize roiSize, IppiMomentState_64f* pCtx)
    IppStatus ippiGetSpatialMoment_64f(IppiMomentState_64f* pState,
                                       int mOrd, int nOrd, int nChannel,
                                       IppiPoint roiOffset, Ipp64f* pValue)
    IppStatus ippiGetCentralMoment_64f(IppiMomentState_64f* pState,
                                       int mOrd, int nOrd, int nChannel,
                                       Ipp64f* pValue)
    IppStatus ippiGetNormalizedSpatialMoment_64f(IppiMomentState_64f* pState,
                                                 int mOrd, int nOrd, int nChannel,
                                                 IppiPoint roiOffset, Ipp64f* pValue)
    IppStatus ippiGetNormalizedCentralMoment_64f(IppiMomentState_64f* pState,
                                                 int mOrd, int nOrd, int nChannel,
                                                 Ipp64f* pValue)
    IppStatus ippiGetHuMoments_64f(IppiMomentState_64f* pState,
                                   int nChannel, IppiHuMoment_64f pHm)

    IppStatus ippiHistogramEven_8u_C1R(Ipp8u* pSrc, int srcStep, IppiSize roiSize,
                                       Ipp32s* pHist, Ipp32s* pLevels,
                                       int nLevels, Ipp32s lowerLevel, Ipp32s upperLevel)

#cdef extern from "ippcv.h":
    IppStatus ippiAdd_8u32f_C1IR(Ipp8u*  pSrc, int srcStep,
                                 Ipp32f* pSrcDst, int srcDstStep,
                                 IppiSize roiSize )
    IppStatus ippiAddSquare_8u32f_C1IR(Ipp8u*  pSrc, int srcStep,
                                       Ipp32f* pSrcDst, int srcDstStep,
                                       IppiSize roiSize)
    IppStatus ippiAddWeighted_8u32f_C1IR( Ipp8u*  pSrc, int srcStep,
                                          Ipp32f* pSrcDst, int srcDstStep,
                                          IppiSize roiSize, Ipp32f alpha )
    IppStatus ippiAddWeighted_32f_C1IR(Ipp32f* pSrc, int srcStep,
                                       Ipp32f* pSrcDst, int srcDstStep,
                                       IppiSize roiSize, Ipp32f alpha )
    IppStatus ippiMean_StdDev_8u_C1R(Ipp8u* pSrc, int srcStep,
                                     IppiSize roiSize,
                                     Ipp64f* mean, Ipp64f* stddev )
    IppStatus ippiMinMaxIndx_8u_C1R( Ipp8u* pSrc, int srcStep, IppiSize roiSize,
                                     Ipp32f* minVal, Ipp32f* maxVal,
                                     IppiPoint* minIdx, IppiPoint* maxIdx )
    IppStatus ippiAbsDiff_8u_C1R( Ipp8u* pSrc1, int srcStep,
                                  Ipp8u* pSrc2, int src2Step,
                                  Ipp8u* pDst, int dstStep, IppiSize roiSize )
    IppStatus ippiAbsDiff_32f_C1R(Ipp32f* pSrc1, int srcStep,
                                  Ipp32f* pSrc2, int src2Step,
                                  Ipp32f* pDst, int dstStep, IppiSize roiSize )
    IppStatus ippiAbs_32f_C1R(Ipp32f* pSrc, int srcStep,
                              Ipp32f* pDst, int dstStep,
                              IppiSize roiSize)
    IppStatus ippiAbs_32f_C1IR(Ipp32f* pSrcDst, int srcDstStep,
                               IppiSize roiSize)
    IppStatus ippiFilterLowpass_8u_C1R(Ipp8u* pSrc, int srcStep,Ipp8u* pDst, int dstStep,IppiSize roiSize,IppiMaskSize maskSize)
    IppStatus ippiFilterSobelHoriz_32f_C1R( Ipp32f *pSrc, int srcStep, Ipp32f *pDst, int dstStep, IppiSize dstRoiSize )
    IppStatus ippiFilterSobelVert_32f_C1R ( Ipp32f *pSrc, int srcStep, Ipp32f *pDst, int dstStep, IppiSize dstRoiSize )
    IppStatus ippiFilterSobelHoriz_8u_C1R( Ipp8u *pSrc, int srcStep, Ipp8u *pDst, int dstStep, IppiSize dstRoiSize )
    IppStatus ippiFilterSobelVert_8u_C1R ( Ipp8u *pSrc, int srcStep, Ipp8u *pDst, int dstStep, IppiSize dstRoiSize )

    IppStatus ippiCFAToRGB_8u_C1C3R(Ipp8u *pSrc, IppiRect srcRoi, IppiSize srcSize, int srcStep, Ipp8u *pDst, int dstStep, IppiBayerGrid bayerGrid, int interpolation)
    IppStatus ippiColorToGray_8u_C3C1R(Ipp8u *pSrc, int srcStep, Ipp8u *pDst, int dstStep, IppiSize roiSize, Ipp32f coeffs[3])
    IppStatus ippiRGBToGray_8u_C3C1R(Ipp8u *pSrc, int srcStep, Ipp8u *pDst, int dstStep, IppiSize roiSize)
    IppStatus ippiRGBToHSV_8u_C3R(Ipp8u *pSrc, int srcStep, Ipp8u *pDst, int dstStep, IppiSize roiSize)
    IppStatus ippiMul_8u_C1IRSfs(Ipp8u *pSrc, int srcStep, Ipp8u *pSrcDst, int srcDstStep, IppiSize roiSize, int scaleFactor)
    IppStatus ippiErode3x3_8u_C1IR( Ipp8u* pSrcDst, int srcDstStep, IppiSize roiSize)

#cdef extern from "ipps.h":
    IppStatus ippSetFlushToZero( int value, unsigned int* pUMask )
    IppStatus ippSetDenormAreZeros( int value )
